package br.ufes.inf.decisionelement;

import br.ufes.inf.decisionelement.gana.networkprofile.NetworkProfile;
import br.ufes.inf.decisionelement.gana.networkprofile.goal.NetworkProfileGoalException;
import br.ufes.inf.decisionelement.reactiveforwarding.ReactivePacketProcessor;
import br.ufes.inf.decisionelement.reactiveforwarding.TopologyDisconnectionListener;
import br.ufes.inf.decisionelement.wheelreinventions.expressionparsing.MathExpressionException;
import br.ufes.inf.decisionelement.wheelreinventions.json.exceptions.JsonParsingException;
import java.util.Dictionary;
import java.util.Properties;
import java.util.concurrent.ExecutorService;
import static java.util.concurrent.Executors.newSingleThreadExecutor;
import static org.onlab.util.Tools.get;
import static org.onlab.util.Tools.groupedThreads;
import org.onosproject.cfg.ComponentConfigService;
import org.onosproject.core.ApplicationId;
import org.onosproject.core.CoreService;
import org.onosproject.net.flow.FlowRuleService;
import org.onosproject.net.flowobjective.FlowObjectiveService;
import org.onosproject.net.host.HostService;
import org.onosproject.net.intent.IntentService;
import org.onosproject.net.packet.PacketProcessor;
import org.onosproject.net.packet.PacketService;
import org.onosproject.net.topology.TopologyService;
import org.osgi.service.component.ComponentContext;
import org.osgi.service.component.annotations.Activate;
import org.osgi.service.component.annotations.Component;
import org.osgi.service.component.annotations.Deactivate;
import org.osgi.service.component.annotations.Modified;
import org.osgi.service.component.annotations.Reference;
import org.osgi.service.component.annotations.ReferenceCardinality;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * ONOS application component.
 */
@Component(
        immediate = true,
        service = { AppComponentInterface.class },
        property = {
            "networkGoals=[]",
            "networkPolicies=[]",
            "networkConfigurations=[]",
        }
)
public class AppComponent implements AppComponentInterface {
    private static final String APP_NAME = "br.ufes.inf.decisionelement";
    private ApplicationId appId;

    // Configurable properties that will show up in ONOS's settings.
    private String networkGoals;
    private String networkPolicies;
    private String networkConfigurations;

    private boolean isActive = false;

    private NetworkProfile networkProfile = new NetworkProfile();

    private final Logger log = LoggerFactory.getLogger(getClass());

    @Reference(cardinality = ReferenceCardinality.MANDATORY)
    protected CoreService coreService;

    @Reference(cardinality = ReferenceCardinality.MANDATORY)
    protected FlowObjectiveService flowObjectiveService;

    @Reference(cardinality = ReferenceCardinality.MANDATORY)
    protected FlowRuleService flowRuleService;

    @Reference(cardinality = ReferenceCardinality.MANDATORY)
    protected HostService hostService;

    @Reference(cardinality = ReferenceCardinality.MANDATORY)
    protected IntentService intentService;

    @Reference(cardinality = ReferenceCardinality.MANDATORY)
    protected PacketService packetService;

    @Reference(cardinality = ReferenceCardinality.MANDATORY)
    protected TopologyService topologyService;

    @Reference(cardinality = ReferenceCardinality.MANDATORY)
    protected ComponentConfigService cfgService;

    protected ReactivePacketProcessor reactivePacketProcessor;
    protected TopologyDisconnectionListener topologyDisconnectionListener;
    protected ExecutorService blackHoleExecutor;

    @Activate
    protected void
        activate(ComponentContext context)
                throws JsonParsingException, NetworkProfileGoalException, MathExpressionException {
        modified(context);
        if (coreService != null) {
            appId = coreService.registerApplication(APP_NAME);
        }
        cfgService.registerProperties(getClass());
        this.isActive = true;
        log.info("Started " + this.getClass().getCanonicalName());
        this.parseConfiguration();
        if (
                reactivePacketProcessor == null
                &&
                appId != null
                &&
                hostService != null
                &&
                topologyService != null
                &&
                flowObjectiveService != null
                &&
                packetService != null
                &&
                flowRuleService != null
        ) {
            blackHoleExecutor = newSingleThreadExecutor(groupedThreads("br/ufes/inf/decisionelement",
                                                                       "black-hole-fixer",
                                                                       log));
            topologyDisconnectionListener = new TopologyDisconnectionListener(
                    appId,
                    blackHoleExecutor,
                    hostService,
                    topologyService,
                    flowRuleService);
            reactivePacketProcessor = new ReactivePacketProcessor(
                    appId,
                    hostService,
                    topologyService,
                    flowObjectiveService,
                    packetService);
            packetService.addProcessor(reactivePacketProcessor, PacketProcessor.director(2));
            topologyService.addListener(topologyDisconnectionListener);
            reactivePacketProcessor.requestIntercepts();
        }
    }

    @Deactivate
    protected void deactivate() {
        cfgService.unregisterProperties(getClass(), false);
        this.isActive = false;
        if (packetService != null && reactivePacketProcessor != null) {
            reactivePacketProcessor.withdrawIntercepts();
            packetService.removeProcessor(reactivePacketProcessor);
            reactivePacketProcessor = null;
        }
        if (topologyService != null && topologyDisconnectionListener != null) {
            topologyService.removeListener(topologyDisconnectionListener);
            topologyDisconnectionListener = null;
        }
        if (blackHoleExecutor != null) {
            blackHoleExecutor.shutdown();
            blackHoleExecutor = null;
        }
        if (flowRuleService != null && appId != null) {
            flowRuleService.removeFlowRulesById(appId);
        }
        appId = null;
        log.info("Stopped " + this.getClass().getCanonicalName());
    }

    @Modified
    public void
        modified(ComponentContext context)
                throws JsonParsingException, MathExpressionException, NetworkProfileGoalException {
        Dictionary<?, ?> properties = context != null ? context.getProperties() : new Properties();
        if (context != null) {
            this.networkGoals = get(properties, "networkGoals");
            this.networkPolicies = get(properties, "networkPolicies");
            this.networkConfigurations = get(properties, "networkConfigurations");
        }
        log.info("Reconfigured " + this.getClass().getCanonicalName());
        this.parseConfiguration();
    }

    @Override
    public void
        parseConfiguration()
                throws JsonParsingException, MathExpressionException, NetworkProfileGoalException {
        if (this.networkGoals == null) {
            this.networkGoals = "[]";
        }
        if (this.networkPolicies == null) {
            this.networkPolicies = "[]";
        }
        if (this.networkConfigurations == null) {
            this.networkConfigurations = "[]";
        }
        networkProfile = NetworkProfile.deJson(networkConfigurations, networkPolicies, networkGoals);
        if (this.isActive) {
            System.out.print("");
        }
    }

    @Override
    public void someMethod() {
        log.info("Invoked " + this.getClass().getCanonicalName());
    }

}
