/*
 * Copyright 2020 Adler Neves.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package br.ufes.inf.decisionelement.wheelreinventions.expressionparsing;

import br.ufes.inf.decisionelement.wheelreinventions.CharArrayUtils;
import br.ufes.inf.decisionelement.wheelreinventions.expressionparsing.syntaxtree.MathTree;
import br.ufes.inf.decisionelement.wheelreinventions.expressionparsing.syntaxtree.MathTreeContext;
import br.ufes.inf.decisionelement.wheelreinventions.expressionparsing.tokenization.AdditionOperatorExpressionParser;
import br.ufes.inf.decisionelement.wheelreinventions.expressionparsing.tokenization.ContextualizationExpressionParser;
import br.ufes.inf.decisionelement.wheelreinventions.expressionparsing.tokenization.DivisionOperatorExpressionParser;
import br.ufes.inf.decisionelement.wheelreinventions.expressionparsing.tokenization.LiteralExpressionParser;
import br.ufes.inf.decisionelement.wheelreinventions.expressionparsing.tokenization.MultiplicationOperatorExpressionParser;
import br.ufes.inf.decisionelement.wheelreinventions.expressionparsing.tokenization.ParenthesisExpressionParser;
import br.ufes.inf.decisionelement.wheelreinventions.expressionparsing.tokenization.SubtractionOperatorExpressionParser;
import br.ufes.inf.decisionelement.wheelreinventions.expressionparsing.tokenization.VariableExpressionParser;
import br.ufes.inf.decisionelement.wheelreinventions.positionalcollections.Tuple;
import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

/**
 * No docstring set.
 * @param <T>
 */
public abstract class MathExpressionParser<T> implements MathStateInterface {
    public final T token;
    public MathExpressionParser() {
        token = null;
    }
    public MathExpressionParser(T t) {
        token = t;
    }
    private static final MathExpressionParser<?>[] EXPRESSION_PARSERS = new MathExpressionParser<?>[]{
        new ParenthesisExpressionParser(),
        new LiteralExpressionParser(),
        new VariableExpressionParser(),
        new MultiplicationOperatorExpressionParser(),
        new DivisionOperatorExpressionParser(),
        new SubtractionOperatorExpressionParser(),
        new AdditionOperatorExpressionParser(),
        new ContextualizationExpressionParser(),
    };
    protected final boolean charSequenceSpans = getCharsSequenceSpans();
    protected final char[] recognizedChars = getRecognizedChars();
    protected abstract boolean getCharsSequenceSpans();
    protected abstract char[] getRecognizedChars();
    @Override
    public boolean isToken() {
        return true;
    }
    @Override
    public boolean isTree() {
        return false;
    }
    @Override
    public boolean isConsumable() {
        return true;
    }
    protected Tuple<String, String> attemptDefaultTokenization(String expression) {
        expression = expression.stripLeading();
        if (expression.length() > 0 && CharArrayUtils.contains(recognizedChars, expression.charAt(0))) {
            int i = 1;
            if (!charSequenceSpans) {
                i--;
            } else {
                for (; i < expression.length(); i++) {
                    if (!CharArrayUtils.contains(recognizedChars, expression.charAt(i))) {
                        i--;
                        break;
                    }
                }
                if (i == expression.length()) {
                    i--;
                }
            }
            return new Tuple<>(expression.substring(0, i + 1).strip(), expression.substring(i + 1).stripLeading());
        }
        return null;
    }
    protected abstract Tuple<MathExpressionParser<?>, String>
        tryTokenization(String expression);
    protected abstract Tuple<List<MathStateInterface>, Integer>
        parseTokens(List<MathStateInterface> tokens, int i)
            throws MathExpressionException;
    public static Tuple<List<MathExpressionParser<?>>, String>
        tokenizePartial(final String expression)
            throws MathExpressionException {
        final List<MathExpressionParser<?>> tokens = new ArrayList<>();
        String remainingExpression = expression.stripLeading();
        boolean tokenFound = true;
        while (tokenFound) {
            tokenFound = false;
            for (final var expressionParser: EXPRESSION_PARSERS) {
                final Tuple<MathExpressionParser<?>, String> resp = expressionParser.
                        tryTokenization(remainingExpression);
                if (resp != null && resp.first != null) {
                    tokenFound = true;
                    remainingExpression = resp.second.stripLeading();
                    tokens.add(resp.first);
                    break;
                }
            }
        }
        if (remainingExpression.stripLeading().length() > 0) {
            throw new MathExpressionException("Cannot tokenize '" + remainingExpression + "'");
        }
        return new Tuple<>(tokens, remainingExpression.stripLeading());
    }
    public static List<MathExpressionParser<?>>
        tokenize(final String expression)
            throws MathExpressionException {
        return tokenizePartial(expression).first;
    }
    public static MathTreeContext
        makeTree(final String expression)
            throws MathExpressionException {
        return makeTree(tokenize(expression));
    }
    public static MathTreeContext
        makeTree(final List<MathExpressionParser<?>> tokens)
            throws MathExpressionException {
        return makeTreePartial(tokens.stream().collect(Collectors.toList()));
    }
    public static MathTreeContext
        makeTreePartial(List<MathStateInterface> tokens)
            throws MathExpressionException {
        if (tokens.size() <= 0) {
            throw new MathExpressionException("Empty expression or sub-expression found");
        }
        for (final var expressionParser: EXPRESSION_PARSERS) {
            int i = 0;
            for (; i < tokens.size(); i++) {
                var token = tokens.get(i);
                if (token instanceof MathExpressionParser) {
                    if (expressionParser.getClass().isInstance(token)) {
                        MathExpressionParser<?> token2 = (MathExpressionParser) token;
                        Tuple<List<MathStateInterface>, Integer> t;
                        t = token2.parseTokens(tokens, i);
                        if (t != null) {
                            tokens = t.first;
                            i = t.second;
                        }
                    }
                }
            }
        }
        if (tokens.size() != 1) {
            throw new MathExpressionException("Tree does not have a root");
        }
        var root2 = tokens.get(0);
        if (root2 instanceof MathTree) {
            var root = (MathTree<?>) root2;
            if (root instanceof MathTreeContext) {
                return (MathTreeContext) root;
            } else {
                return new MathTreeContext(
                        new Tuple<>(
                                new ArrayList<>(),
                                root
                        )
                );
            }
        } else {
            throw new MathExpressionException("Tree does not have a root");
        }
    }
    @Override
    public String toString() {
        return "" + token;
    }
}
