/*
 * Copyright 2020 Adler Neves.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package br.ufes.inf.decisionelement.wheelreinventions.json;

import br.ufes.inf.decisionelement.wheelreinventions.json.exceptions.JsonParsingException;
import br.ufes.inf.decisionelement.wheelreinventions.json.exceptions.JsonParsingUnexpectedEndException;
import br.ufes.inf.decisionelement.wheelreinventions.json.exceptions.JsonParsingUnexpectedSymbolException;
import br.ufes.inf.decisionelement.wheelreinventions.positionalcollections.Tuple;
import javax.validation.constraints.NotNull;

/**
 * No docstring set.
 * @param <T>
 */
public abstract class JsonElement<T> {
    protected T representation;
    public JsonElement() {
        representation = null;
    }
    public JsonElement(T t) {
        representation = t;
    }
    public T getRepresentation() {
        return representation;
    }
    protected char[] handleChars = getHandleChars();
    public static @NotNull JsonElement
        parseJson(@NotNull String json)
                throws JsonParsingException {
        return parseJsonPartial(json).first;
    }
    protected boolean handlesParsing(char c) {
        for (char hc: handleChars) {
            if (hc == c) {
                return true;
            }
        }
        return false;
    };
    protected abstract char[] getHandleChars();
    protected abstract Tuple<JsonElement<?>, String> parsePartial(String json) throws JsonParsingException;
    protected static void parsePartialThrower(String json) throws JsonParsingUnexpectedSymbolException {
        throw new JsonParsingUnexpectedSymbolException(
                "Unexpected character for JSON: \'" + json.charAt(0) + '\''
        );
    }
    protected static UnsupportedOperationException parseFinalExcGen() {
        return new UnsupportedOperationException("Not supported yet.");
    }
    public static @NotNull Tuple<JsonElement<?>, String>
        parseJsonPartial(@NotNull String json)
                throws JsonParsingException {
        json = json.stripLeading();
        if (json.length() <= 0) {
            throw new JsonParsingUnexpectedEndException("Cannot parse JSON from empty string.");
        } else {
            for (JsonElement<?> pho : PARSE_HANDLER_OBJECTS) {
                if (pho.handlesParsing(json.charAt(0))) {
                    return pho.parsePartial(json);
                }
            }
            throw parseFinalExcGen();
        }
    }
    private static final JsonElement<?>[] PARSE_HANDLER_OBJECTS = new JsonElement<?>[]{
        new JsonArray(),
        new JsonObject(),
        new JsonString(),
        new JsonNumeric(),
        new JsonBoolean(),
        new JsonNull(),
    };
    public JsonObject castToObject() {
        if (this instanceof JsonObject) {
            return (JsonObject) this;
        } else {
            throw new ClassCastException(
                    "Cannot convert " +
                    this.getClass().getSimpleName() +
                    " to JsonObject");
        }
    }
    public JsonArray castToArray() {
        if (this instanceof JsonArray) {
            return (JsonArray) this;
        } else {
            throw new ClassCastException(
                    "Cannot convert " +
                    this.getClass().getSimpleName() +
                    " to JsonArray");
        }
    }
    public JsonBoolean castToBoolean() {
        if (this instanceof JsonBoolean) {
            return (JsonBoolean) this;
        } else {
            throw new ClassCastException(
                    "Cannot convert " +
                    this.getClass().getSimpleName() +
                    " to JsonBoolean");
        }
    }
    public JsonNumeric castToNumeric() {
        if (this instanceof JsonNumeric) {
            return (JsonNumeric) this;
        } else {
            throw new ClassCastException(
                    "Cannot convert " +
                    this.getClass().getSimpleName() +
                    " to JsonNumeric");
        }
    }
    public JsonNull castToNull() {
        if (this instanceof JsonNull) {
            return (JsonNull) this;
        } else {
            throw new ClassCastException(
                    "Cannot convert " +
                    this.getClass().getSimpleName() +
                    " to JsonNull");
        }
    }
    public boolean isJsonObject() {
        return this instanceof JsonObject;
    }
    public boolean isJsonArray() {
        return this instanceof JsonArray;
    }
    public boolean isJsonBoolean() {
        return this instanceof JsonBoolean;
    }
    public boolean isJsonNumeric() {
        return this instanceof JsonNumeric;
    }
    public boolean isJsonNull() {
        return this instanceof JsonNull;
    }
}
