/*
* Copyright 2020 Adler Neves.
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/
package br.ufes.inf.decisionelement.wheelreinventions.json;

import static br.ufes.inf.decisionelement.wheelreinventions.json.JsonElement.parseJsonPartial;
import br.ufes.inf.decisionelement.wheelreinventions.json.exceptions.JsonParsingException;
import br.ufes.inf.decisionelement.wheelreinventions.json.exceptions.JsonParsingUnexpectedEndException;
import br.ufes.inf.decisionelement.wheelreinventions.json.exceptions.JsonParsingUnexpectedTokenException;
import br.ufes.inf.decisionelement.wheelreinventions.positionalcollections.Tuple;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 * No docstring set.
 * @param <T>
 */
public class JsonObject extends JsonElement<Map<String, JsonElement>> implements Map<String, JsonElement> {
    public JsonObject() {
        super(new HashMap<>());
    }
    public JsonObject(Map<String, JsonElement> m) {
        super(m);
    }
    @Override
    public String toString() {
        final Object[] keys = representation.keySet().toArray();
        Arrays.sort(keys);
        final var kl = keys.length;
        final var sb = new StringBuilder();
        sb.append('{');
        for (int i = 0; i < kl; i++) {
            final var key = keys[i].toString();
            final var value = representation.get(key);
            sb.append((new JsonString(key)).toString());
            sb.append(':');
            sb.append(value.toString());
            if ((i + 1) < kl) {
                sb.append(',');
            }
        }
        sb.append('}');
        return sb.toString();
    }
    @Override
    protected char[] getHandleChars() {
        return new char[]{'{'};
    }
    @Override
    protected Tuple<JsonElement<?>, String> parsePartial(String json) throws JsonParsingException {
        switch (json.charAt(0)) {
            case '{': // object
            {
                var o = new JsonObject();
                String jsonRemaining = json;
                while (true) {
                    if (jsonRemaining.length() <= 0) {
                        throw new JsonParsingUnexpectedEndException("Unterminated JSON object");
                    } else {
                        boolean opens = false;
                        switch (jsonRemaining.charAt(0)) {
                            case '{':
                                opens = true;
                            case ',':
                                jsonRemaining = jsonRemaining.substring(1).stripLeading();
                                if (opens && jsonRemaining.length() > 0 && jsonRemaining.charAt(0) == '}') {
                                    break;
                                }
                                var partialResult = parseJsonPartial(jsonRemaining);
                                JsonString js = (JsonString) partialResult.first;
                                jsonRemaining = partialResult.second.stripLeading();
                                if (jsonRemaining.length() <= 0) {
                                    throw new JsonParsingUnexpectedEndException(
                                            "Unterminated JSON object was expecting token: ':'"
                                    );
                                }
                                if (jsonRemaining.charAt(0) != ':') {
                                    throw new JsonParsingUnexpectedTokenException(
                                            "Expected object's key-value separator token: ':'"
                                    );
                                }
                                partialResult = parseJsonPartial(jsonRemaining.substring(1));
                                o.representation.put(
                                        js.representation,
                                        partialResult.first
                                );
                                jsonRemaining = partialResult.second.stripLeading();
                                break;
                            case '}':
                                return new Tuple<>(
                                        o,
                                        jsonRemaining.substring(1)
                                );
                            default:
                                throw new JsonParsingUnexpectedTokenException(
                                        "Expected object token: ',' or '}'"
                                );
                        }
                    }
                }
            }
            default: // unknown
                parsePartialThrower(json);
        }
        throw parseFinalExcGen();
    }

    @Override
    public int size() {
        return representation.size();
    }

    @Override
    public boolean isEmpty() {
        return representation.isEmpty();
    }

    @Override
    public boolean containsKey(Object arg0) {
        return representation.containsKey(arg0);
    }

    @Override
    public boolean containsValue(Object arg0) {
        return representation.containsValue(arg0);
    }

    @Override
    public JsonElement get(Object arg0) {
        return representation.get(arg0);
    }

    @Override
    public JsonElement put(String arg0, JsonElement arg1) {
        return representation.put(arg0, arg1);
    }

    @Override
    public JsonElement remove(Object arg0) {
        return representation.remove(arg0);
    }

    @Override
    public void putAll(Map<? extends String, ? extends JsonElement> arg0) {
        representation.putAll(arg0);
    }

    @Override
    public void clear() {
        representation.clear();
    }

    @Override
    public Set<String> keySet() {
        return representation.keySet();
    }

    @Override
    public Collection<JsonElement> values() {
        return representation.values();
    }

    @Override
    public Set<Entry<String, JsonElement>> entrySet() {
        return representation.entrySet();
    }
}
