/*
* Copyright 2020 Adler Neves.
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/
package br.ufes.inf.decisionelement.wheelreinventions.json;

import br.ufes.inf.decisionelement.wheelreinventions.json.exceptions.JsonParsingException;
import br.ufes.inf.decisionelement.wheelreinventions.json.exceptions.JsonParsingUnexpectedSymbolException;
import br.ufes.inf.decisionelement.wheelreinventions.positionalcollections.Tuple;

/**
 * No docstring set.
 */
public class JsonString extends JsonElement<String> {
    public JsonString() {
        super("");
    }
    public JsonString(String s) {
        super(s);
    }
    @Override
    public String toString() {
        final var sb = new StringBuilder();
        sb.append('"');
        sb.append(representation
                .replaceAll("\\\\", "\\\\\\\\") // "\" → "\\"
                .replaceAll("\n", "\\\\n") // \n → "\\n"
                .replaceAll("\r", "\\\\r") // \r → "\\r"
                .replaceAll("\t", "\\\\t") // \t → "\\t"
                .replaceAll("\"", "\\\\\"") // \" → "\\""
                .replaceAll("\'", "\\\\\'")); // \' → "\\'"
        sb.append('"');
        return sb.toString();
    }
    @Override
    protected char[] getHandleChars() {
        return new char[]{'"'};
    }
    @Override
    protected Tuple<JsonElement<?>, String> parsePartial(String json) throws JsonParsingException {
        switch (json.charAt(0)) {
            case '"': // string
            {
                boolean stringEnded = false;
                boolean escapeSequence = false;
                var sb = new StringBuilder();
                int i = 1;
                for (; i < json.length(); i++) {
                    char c = json.charAt(i);
                    if (!escapeSequence) {
                        switch (c) {
                            case '\\':
                                escapeSequence = true;
                                break;
                            case '"':
                                stringEnded = true;
                                i++;
                                break;
                            default:
                                sb.append(c);
                        }
                    } else {
                        escapeSequence = false;
                        switch (c) {
                            case 't':
                                sb.append('\t');
                                break;
                            case 'r':
                                sb.append('\r');
                                break;
                            case 'n':
                                sb.append('\n');
                                break;
                            case 'b':
                                sb.append('\b');
                                break;
                            case '0':
                                sb.append('\0');
                                break;
                            case 'u':
                                int codepoint = Integer.parseInt(json.substring(i + 1, i + 5), 16);
                                sb.append(Character.toChars(codepoint));
                                i += 4;
                                break;
                            default:
                                sb.append(c);
                        }
                    }
                    if (stringEnded) {
                        break;
                    }
                }
                if (!stringEnded) {
                    throw new JsonParsingUnexpectedSymbolException("Unterminated JSON string");
                }
                var s = new JsonString();
                s.representation = sb.toString();
                return new Tuple<>(s, json.substring(i));
            }
            default: // unknown
                parsePartialThrower(json);
        }
        throw parseFinalExcGen();
    }
}
